"""
Utility class for working with temporary files
"""

import os

from tempfile import mkstemp


class MixedModeTemporaryFile:
    """
    Temporary file that can be opened multiple times in different modes
    """

    def __init__(self):
        self._fd, self._filename = mkstemp()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, traceback):
        self.close()

    def open(self, mode):
        """
        Open the file in the specified mode
        """
        return os.fdopen(self._fd, mode, closefd=False)

    def close(self):
        """
        Close the temporary file (and delete it)
        """
        os.close(self._fd)
        os.unlink(self._filename)

    def rewind(self):
        """
        Seek to the beginning of the file
        """
        os.lseek(self._fd, 0, os.SEEK_SET)

    def size(self):
        """
        Return the size of the file
        """
        return os.stat(self._fd).st_size



from zipfile import ZipFile


class ZipArchiver:
    """
    Archiver for creating ZIP files
    """

    # pylint: disable=unused-argument
    @staticmethod
    def get_extension(fmt):
        """
        Determine the correct file extension (always .zip)
        """
        return 'zip'

    # pylint: disable=unused-argument
    def __init__(self, fileobj, fmt):
        self._fileobj = fileobj

    def __enter__(self):
        # pylint: disable=attribute-defined-outside-init
        self._zipfile = ZipFile(
            file=self._fileobj,
            mode='w',
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._zipfile.close()

    def add(self, filename, size, fileobj):
        """
        Add the provided file to the archive
        """
        # Loop through the contents of the fileobj,
        # writing them to the zipfile in chunks
        bytes_remaining = size
        with self._zipfile.open(filename, 'w') as cfile:
            while bytes_remaining:
                chunk_size = min(65536, bytes_remaining)
                cfile.write(fileobj.read(chunk_size))
                bytes_remaining -= chunk_size



from .tarball import TarballArchiver
from .zipfile import ZipArchiver


TARBALL = TarballArchiver.UNCOMPRESSED
TARBALL_GZ = TarballArchiver.GZ
TARBALL_BZ2 = TarballArchiver.BZ2
TARBALL_XZ = TarballArchiver.XZ
ZIP = 'zip'


def get_archiver(fmt):
    """
    Return the class corresponding with the provided archival format
    """
    if fmt in (TARBALL, TARBALL_GZ, TARBALL_BZ2, TARBALL_XZ):
        return TarballArchiver
    if fmt == ZIP:
        return ZipArchiver
    raise KeyError("Invalid format '{}' specified".format(fmt))



from tarfile import TarInfo, TarFile


class TarballArchiver:
    """
    Archiver for creating tarballs (compressed and uncompressed)
    """

    UNCOMPRESSED = 'tar'
    GZ = 'gz'
    BZ2 = 'bz2'
    XZ = 'xz'

    @classmethod
    def get_extension(cls, fmt):
        """
        Determine the correct file extension for the archive
        """
        return fmt if fmt == cls.UNCOMPRESSED else 'tar.{}'.format(fmt)

    def __init__(self, fileobj, fmt):
        self._fileobj = fileobj
        self._fmt = fmt

    def __enter__(self):
        # pylint: disable=attribute-defined-outside-init
        self._tarfile = TarFile.open(
            mode='w:{}'.format(self._fmt),
            fileobj=self._fileobj,
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._tarfile.close()

    def add(self, filename, size, fileobj):
        """
        Add the provided file to the archive
        """
        tarinfo = TarInfo(filename)
        tarinfo.size = size
        self._tarfile.addfile(tarinfo, fileobj)



"""
Create an archive
"""

from contextlib import contextmanager
from datetime import datetime
from json import dump
from os import path

from django.apps import apps
from django.conf import settings
from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.db import connection, models
from django.db.migrations.recorder import MigrationRecorder

from ... import __version__
from ...archivers import get_archiver, TARBALL_BZ2
from ...util.file import MixedModeTemporaryFile


class Command(BaseCommand):
    """
    Create a compressed archive of database tables and uploaded media.
    """

    help = "Create a compressed archive of database tables and uploaded media."

    _DEFAULT_ARCHIVE_EXCLUDE = (
        'auth.Permission',
        'contenttypes.ContentType',
        'sessions.Session',
    )

    @staticmethod
    def _get_filename(archiver, fmt):
        return path.join(
            getattr(settings, 'ARCHIVE_DIRECTORY', ''),
            "{}.{}".format(
                datetime.today().strftime(
                    getattr(
                        settings,
                        'ARCHIVE_FILENAME',
                        '%Y-%m-%d--%H-%M-%S',
                    ),
                ),
                archiver.get_extension(fmt),
            ),
        )

    @staticmethod
    @contextmanager
    def _write_to_archive(archive, filename):
        with MixedModeTemporaryFile() as tempfile:
            with tempfile.open('w') as fileobj:
                yield fileobj
            tempfile.rewind()
            with tempfile.open('rb') as fileobj:
                archive.add(filename, tempfile.size(), fileobj)

    @staticmethod
    def _dump_meta(archive):
        with Command._write_to_archive(archive, 'meta.json') as fileobj:
            dump(
                {
                    'version': __version__,
                    'migrations': dict(
                        MigrationRecorder(connection)
                        .applied_migrations()
                        .keys()
                    ),
                },
                fileobj,
                indent=2,
            )

    @staticmethod
    def _dump_db(archive):
        with Command._write_to_archive(archive, 'data.json') as fileobj:
            call_command(
                'dumpdata',
                all=True,
                format='json',
                exclude=getattr(
                    settings,
                    'ARCHIVE_EXCLUDE',
                    Command._DEFAULT_ARCHIVE_EXCLUDE,
                ),
                stdout=fileobj,
            )

    @staticmethod
    def _dump_files(archive):
        for app_config in apps.get_app_configs():
            for model in app_config.get_models():
                field_names = []
                for field in model._meta.fields:
                    if isinstance(field, models.FileField):
                        field_names.append(field.name)
                if not field_names:
                    continue
                for row in model.objects.all():
                    for field_name in field_names:
                        field = getattr(row, field_name)
                        if field:
                            with field as field:
                                archive.add(field.name, field.size, field)

    # pylint: disable=unused-argument
    def handle(self, *args, **kwargs):
        """
        Process the command
        """
        fmt = getattr(settings, 'ARCHIVE_FORMAT', TARBALL_BZ2)
        archiver = get_archiver(fmt)
        filename = Command._get_filename(archiver, fmt)
        with open(filename, 'wb') as fileobj:
            archive = archiver(fileobj, fmt)
            with archive:
                Command._dump_meta(archive)
                Command._dump_db(archive)
                Command._dump_files(archive)
        self.stdout.write(
            self.style.SUCCESS(
                "Successfully wrote {}".format(filename),
            ),
        )
